@@ -6,8 +6,8 @@
 
       Jane Street Holding, LLC
       Author: Markus Mottl
-      email: mmottl\@janestcapital.com
-      WWW: http://www.janestcapital.com/ocaml
+      email: mmottl\@janestreet.com
+      WWW: http://www.janestreet.com/ocaml
 
    This file is derived from file "pa_tywith.ml" of version 0.45 of the
    library "Tywith".
@@ -80,7 +80,7 @@ let unroll_fun_matches _loc fp1 fp2 =
   | _ -> assert false  (* impossible *)
 
 let rec sig_of_tds cnv = function
-  | TyDcl (_loc, type_name, tps, _rhs, _cl) -> cnv _loc type_name tps
+  | TyDcl (_loc, type_name, tps, rhs, cl) -> cnv _loc type_name tps rhs cl
   | TyAnd (_loc, tp1, tp2) ->
       <:sig_item< $sig_of_tds cnv tp1$; $sig_of_tds cnv tp2$ >>
   | _ -> assert false  (* impossible *)
@@ -90,7 +90,7 @@ let rec sig_of_tds cnv = function
 
 (* Generates the signature for type conversion to S-expressions *)
 module Sig_generate_sexp_of = struct
-  let sig_of_td _loc type_name tps =
+  let sig_of_td _loc type_name tps _rhs _cl =
     let rec loop this_type = function
       | [] -> <:ctyp< $this_type$ -> Sexplib.Sexp.t >>
       | tp :: tps ->
@@ -119,7 +119,7 @@ end
 
 (* Generates the signature for type conversion from S-expressions *)
 module Sig_generate_of_sexp = struct
-  let sig_of_td with_poly _loc type_name tps =
+  let sig_of_td with_poly _loc type_name tps rhs _cl =
     let rec loop this_tp = function
       | [] -> <:ctyp< Sexplib.Sexp.t -> $this_tp$ >>
       | tp :: tps ->
@@ -131,12 +131,38 @@ module Sig_generate_of_sexp = struct
     let of_sexp_item =
       <:sig_item< value $lid: type_name ^ "_of_sexp"$ : $of_sexp$; >>
     in
-    if with_poly then
-      <:sig_item<
-        $of_sexp_item$;
-        value $lid: type_name ^ "_of_sexp__"$ : $of_sexp$;
-      >>
-    else of_sexp_item
+    let is_polymorphic_variant =
+      let rec loop = function
+        | <:ctyp< private $tp$ >> -> loop tp
+        | <:ctyp< ( $tup:_$ ) >>
+        | <:ctyp< $_$ -> $_$ >>
+        | <:ctyp< { $_$ } >>
+        | <:ctyp< [ $_$ ] >> -> `Surely_not
+        | <:ctyp< [< $_$ ] >> | <:ctyp< [> $_$ ] >>
+        | <:ctyp< [= $_$ ] >> -> `Definitely
+        | <:ctyp< '$_$ >>
+        | <:ctyp< $_$ $_$ >>
+        | <:ctyp< $id:_$ >>
+        | <:ctyp< >> -> `Maybe
+        | <:ctyp< $tp1$ == $tp2$ >> ->
+            begin match loop tp1 with
+            | (`Surely_not | `Definitely) as res -> res
+            | `Maybe -> loop tp2 end
+        | _ -> failwith "Sig_generate_of_sexp.sig_of_td: unknown type"
+      in
+      loop rhs
+    in
+    match with_poly, is_polymorphic_variant with
+    | true, `Surely_not ->
+        failwith
+          "Sig_generate_of_sexp.sig_of_td: sexp_poly annotation \
+          but type is surely not a polymorphic variant"
+    | false, (`Surely_not | `Maybe) -> of_sexp_item
+    | (true | false), `Definitely | true, `Maybe ->
+        <:sig_item<
+          $of_sexp_item$;
+          value $lid: type_name ^ "_of_sexp__"$ : $of_sexp$;
+        >>
 
   let mk_sig with_poly tds =
     <:sig_item< $sig_of_tds (sig_of_td with_poly) tds$ >>
@@ -200,6 +226,7 @@ module Generate_sexp_of = struct
     | ["mat"] -> <:expr< Sexplib.Conv.sexp_of_mat >>
     | ["float32_mat"] -> <:expr< Sexplib.Conv.sexp_of_float32_mat >>
     | ["float64_mat"] -> <:expr< Sexplib.Conv.sexp_of_float64_mat >>
+    | ["exn"] -> <:expr< Sexplib.Conv.sexp_of_exn >>
     | tn :: rev_path -> mk_abst_call _loc tn rev_path
     | [] -> assert false  (* impossible *)
 
@@ -238,11 +265,29 @@ module Generate_sexp_of = struct
 
   (* Conversion of variant types *)
 
+  and mk_cnv_expr _loc tp =
+    match sexp_of_type _loc tp with
+    | `Fun fun_expr -> <:expr< $fun_expr$ >>
+    | `Match matchings -> <:expr< fun el -> match el with [ $matchings$ ] >>
+
   and sexp_of_variant _loc row_fields =
     let rec loop = function
       | <:ctyp< $tp1$ | $tp2$ >> -> <:match_case< $loop tp1$ | $loop tp2$ >>
       | <:ctyp< `$cnstr$ >> ->
           <:match_case< `$cnstr$ -> Sexplib.Sexp.Atom $str:cnstr$ >>
+      | <:ctyp< `$cnstr$ of sexp_list $tp$>> ->
+        let cnv_expr =
+          match sexp_of_type _loc tp with
+          | `Fun fun_expr -> <:expr< $fun_expr$ >>
+          | `Match matchings ->
+              <:expr< fun el -> match el with [ $matchings$ ] >>
+        in
+        <:match_case<
+          `$cnstr$ l ->
+             Sexplib.Sexp.List
+               [ Sexplib.Sexp.Atom $str:cnstr$ ::
+                   Sexplib.Conv.list_map $cnv_expr$ l]
+        >>
       | <:ctyp< `$cnstr$ of $tps$ >> ->
           let fps = List.map ~f:(sexp_of_type _loc) (list_of_ctyp tps []) in
           let bindings, patts, vars = mk_bindings _loc fps in
@@ -254,8 +299,8 @@ module Generate_sexp_of = struct
             >>
           in
           <:match_case< `$cnstr$ $paSem_of_list patts$ -> $expr$ >>
-      | <:ctyp< [= $row_fields$ ] >> | <:ctyp< [> $row_fields$ ] >>
-      | <:ctyp< [< $row_fields$ ] >> -> loop row_fields
+      | <:ctyp< [< $row_fields$ ] >> | <:ctyp< [> $row_fields$ ] >>
+      | <:ctyp< [= $row_fields$ ] >> -> loop row_fields
       | <:ctyp< $tp1$ $tp2$ >> ->
           let id_path = Gen.get_appl_path _loc tp1 in
           let call = sexp_of_appl_fun _loc tp1 tp2 in
@@ -302,6 +347,19 @@ module Generate_sexp_of = struct
         <:match_case< $branch_sum _loc tp1$ | $branch_sum _loc tp2$ >>
     | <:ctyp< $uid:cnstr$ >> ->
         <:match_case< $uid:cnstr$ -> Sexplib.Sexp.Atom $str:cnstr$ >>
+    | <:ctyp< $uid:cnstr$ of sexp_list $tp$>> ->
+        let cnv_expr =
+          match sexp_of_type _loc tp with
+          | `Fun fun_expr -> <:expr< $fun_expr$ >>
+          | `Match matchings ->
+              <:expr< fun el -> match el with [ $matchings$ ] >>
+        in
+        <:match_case<
+          $uid:cnstr$ l ->
+             Sexplib.Sexp.List
+               [Sexplib.Sexp.Atom $str:cnstr$ ::
+                   Sexplib.Conv.list_map $cnv_expr$ l]
+        >>
     | <:ctyp< $uid:cnstr$ of $tps$ >> ->
         let fps = List.map ~f:(sexp_of_type _loc) (list_of_ctyp tps []) in
         let cnstr_expr = <:expr< Sexplib.Sexp.Atom $str:cnstr$ >> in
@@ -327,10 +385,30 @@ module Generate_sexp_of = struct
     let p = <:patt< $lid:name$ = $lid:"v_" ^ name$ >> in
     <:patt< $patt$; $p$ >>
 
-  let mk_cnv_expr _loc tp var =
-    match sexp_of_type _loc tp with
-    | `Fun fun_expr -> <:expr< $fun_expr$ $var$ >>
-    | `Match matchings -> <:expr< match $var$ with [ $matchings$ ] >>
+  let sexp_of_default_field _loc patt expr name tp sexp_of empty =
+    let patt = mk_rec_patt _loc patt name in
+    let cnv_expr =
+      match sexp_of_type _loc tp with
+      | `Fun fun_expr -> <:expr< $fun_expr$ >>
+      | `Match matchings ->
+          <:expr< fun el -> match el with [ $matchings$ ] >>
+    in
+    let expr =
+      let v_name = <:expr< $lid: "v_" ^ name$ >> in
+      <:expr<
+        let bnds =
+          if $v_name$ = $empty$ then bnds
+          else
+            let arg = $sexp_of$ $cnv_expr$ $v_name$ in
+            let bnd =
+              Sexplib.Sexp.List [Sexplib.Sexp.Atom $str:name$; arg]
+            in
+            [ bnd :: bnds ]
+        in
+        $expr$
+      >>
+    in
+    patt, expr
 
   let sexp_of_record _loc flds_ctyp =
     let flds = list_of_ctyp flds_ctyp [] in
@@ -358,29 +436,12 @@ module Generate_sexp_of = struct
           patt, expr
       | <:ctyp< $lid:name$ : mutable sexp_list $tp$ >>
       | <:ctyp< $lid:name$ : sexp_list $tp$ >> ->
-          let patt = mk_rec_patt _loc patt name in
-          let cnv_expr =
-            match sexp_of_type _loc tp with
-            | `Fun fun_expr -> <:expr< $fun_expr$ >>
-            | `Match matchings ->
-                <:expr< fun el -> match el with [ $matchings$ ] >>
-          in
-          let expr =
-            let v_name = <:expr< $lid: "v_" ^ name$ >> in
-            <:expr<
-              let bnds =
-                if $v_name$ = [] then bnds
-                else
-                  let arg = sexp_of_list $cnv_expr$ $v_name$ in
-                  let bnd =
-                    Sexplib.Sexp.List [Sexplib.Sexp.Atom $str:name$; arg]
-                  in
-                  [ bnd :: bnds ]
-              in
-              $expr$
-            >>
-          in
-          patt, expr
+          sexp_of_default_field
+            _loc patt expr name tp <:expr< sexp_of_list >> <:expr< [] >>
+      | <:ctyp< $lid:name$ : mutable sexp_array $tp$ >>
+      | <:ctyp< $lid:name$ : sexp_array $tp$ >> ->
+          sexp_of_default_field
+            _loc patt expr name tp <:expr< sexp_of_array >> <:expr< [||] >>
       | <:ctyp< $lid:name$ : mutable $tp$ >>
       | <:ctyp< $lid:name$ : $tp$ >> ->
           let patt = mk_rec_patt _loc patt name in
@@ -536,11 +597,11 @@ module Generate_of_sexp = struct
   (* Generate code for matching malformed S-expressions *)
   let mk_variant_other_matches _loc rev_els call =
     let coll_structs acc (_loc, cnstr) =
-      <:match_case< $str:cnstr$ -> Sexplib.Conv_error.$lid:call$ _loc sexp >>
+      <:match_case< $str:cnstr$ -> Sexplib.Conv_error.$lid:call$ _loc _sexp >>
         :: acc
     in
     let exc_no_variant_match =
-      <:match_case< _ -> Sexplib.Conv_error.no_variant_match _loc sexp >>
+      <:match_case< _ -> Sexplib.Conv_error.no_variant_match _loc _sexp >>
     in
     List.fold_left ~f:coll_structs ~init:[exc_no_variant_match] rev_els
 
@@ -563,9 +624,8 @@ module Generate_of_sexp = struct
           ainhs,
           `S (_loc, cnstr, tps) :: sinhs
         )
-    | <:ctyp< [= $row_fields$ ] >>
-    | <:ctyp< [> $row_fields$ ] >>
-    | <:ctyp< [< $row_fields$ ] >> ->
+    | <:ctyp< [< $row_fields$ ] >> | <:ctyp< [> $row_fields$ ] >>
+    | <:ctyp< [= $row_fields$ ] >> ->
         List.fold_left ~f:(split_row_field _loc)
           ~init:acc (list_of_ctyp row_fields [])
     | <:ctyp< $_$ $_$ >>
@@ -617,7 +677,10 @@ module Generate_of_sexp = struct
     | <:ctyp< sexp_option >> ->
         `Fun <:expr< fun a_of_sexp v -> Some (a_of_sexp v) >>
     | <:ctyp< sexp_list >> ->
-        `Fun <:expr< fun a_of_sexp v -> list_of_sexp a_of_sexp v >>
+        `Fun <:expr< fun a_of_sexp v -> Sexplib.Conv.list_of_sexp a_of_sexp v >>
+    | <:ctyp< sexp_array >> ->
+        `Fun <:expr< fun a_of_sexp v ->
+          Sexplib.Conv.array_of_sexp a_of_sexp v >>
     | <:ctyp< $tp1$ $tp2$ >> ->
         let fp1 = type_of_sexp _loc tp1 in
         let fp2 = type_of_sexp _loc tp2 in
@@ -671,7 +734,7 @@ module Generate_of_sexp = struct
     let new_other_matches =
       [
         <:match_case<
-          _ -> try ($fun_expr$ sexp :> $full_type$) with [ $match_exc$ ]
+          _ -> try ($fun_expr$ _sexp :> $full_type$) with [ $match_exc$ ]
         >>
       ]
     in
@@ -698,25 +761,38 @@ module Generate_of_sexp = struct
 
   (* Match arguments of constructors (variants or sum types) *)
   and mk_cnstr_args_match _loc ~is_variant cnstr tps =
-    let fps = List.map ~f:(type_of_sexp _loc) (list_of_ctyp tps []) in
-    let bindings, patts, vars = mk_bindings _loc fps in
-    let good_arg_match_expr =
-      let vars_expr =
-        match vars with
-        | [var_expr] -> var_expr
-        | _ -> <:expr< ( $tup:exCom_of_list vars$ ) >>
-      in
+    let cnstr vars_expr =
       if is_variant then <:expr< `$cnstr$ $vars_expr$ >>
       else <:expr< $uid:cnstr$ $vars_expr$ >>
     in
-    let handle_exc =
-      if is_variant then "ptag_incorrect_n_args" else "stag_incorrect_n_args"
-    in
-    <:expr<
-      match sexp_args with
-      [ $Gen.mk_patt_lst _loc patts$ -> let $bindings$ in $good_arg_match_expr$
-      | _ -> Sexplib.Conv_error.$lid:handle_exc$ _loc tag sexp ]
-    >>
+    match tps with
+    | <:ctyp< sexp_list $tp$ >> ->
+      let cnv =
+        match type_of_sexp _loc tp with
+        | `Fun fun_expr -> <:expr< $fun_expr$ >>
+        | `Match matchings ->
+            <:expr< fun el -> match el with [ $matchings$ ] >>
+      in
+      cnstr <:expr< Sexplib.Conv.list_map ($cnv$) sexp_args >>
+    | _ ->
+        let fps = List.map ~f:(type_of_sexp _loc) (list_of_ctyp tps []) in
+        let bindings, patts, vars = mk_bindings _loc fps in
+        let good_arg_match_expr =
+          let vars_expr =
+            match vars with
+            | [var_expr] -> var_expr
+            | _ -> <:expr< ( $tup:exCom_of_list vars$ ) >>
+          in
+          cnstr vars_expr
+        in
+        let handle_exc =
+          if is_variant then "ptag_incorrect_n_args" else "stag_incorrect_n_args"
+        in
+        <:expr<
+          match sexp_args with
+            [ $Gen.mk_patt_lst _loc patts$ -> let $bindings$ in $good_arg_match_expr$
+            | _ -> Sexplib.Conv_error.$lid:handle_exc$ _loc _tag _sexp ]
+        >>
 
   (* Generate code for matching structured variants *)
   and mk_variant_match_struct _loc full_type rev_structs_inhs rev_atoms =
@@ -725,7 +801,7 @@ module Generate_of_sexp = struct
       | `S (_loc, cnstr, tps) ->
           has_structs_ref := true;
           let expr = mk_cnstr_args_match _loc ~is_variant:true cnstr tps in
-          let new_match = <:match_case< ($str:cnstr$ as tag) -> $expr$ >> in
+          let new_match = <:match_case< ($str:cnstr$ as _tag) -> $expr$ >> in
           new_match :: other_matches, false
       | `I (_loc, inh) ->
           handle_variant_inh _loc full_type match_last other_matches inh
@@ -754,10 +830,10 @@ module Generate_of_sexp = struct
       else <:patt< _ >>
     in
     <:match_case<
-        Sexplib.Sexp.Atom atom as sexp ->
+        Sexplib.Sexp.Atom atom as _sexp ->
           $mk_variant_match_atom _loc full_type rev_atoms_inhs rev_structs$
       | Sexplib.Sexp.List
-          [Sexplib.Sexp.Atom atom :: $maybe_sexp_args_patt$] as sexp ->
+          [Sexplib.Sexp.Atom atom :: $maybe_sexp_args_patt$] as _sexp ->
             $match_struct$
       | Sexplib.Sexp.List [Sexplib.Sexp.List _ :: _] as sexp ->
           Sexplib.Conv_error.nested_list_invalid_poly_var _loc sexp
@@ -847,8 +923,8 @@ module Generate_of_sexp = struct
         let lccnstr = String.uncapitalize cnstr in
         <:match_case<
           (Sexplib.Sexp.List
-            [Sexplib.Sexp.Atom ($str:lccnstr$ | $str:cnstr$ as tag) ::
-              sexp_args] as sexp) ->
+            [Sexplib.Sexp.Atom ($str:lccnstr$ | $str:cnstr$ as _tag) ::
+              sexp_args] as _sexp) ->
                 $mk_cnstr_args_match _loc ~is_variant:false cnstr tps$
         >>
     | <:ctyp< $tp1$ | $tp2$ >> ->
@@ -908,7 +984,7 @@ module Generate_of_sexp = struct
     | <:ctyp< $lid:nm$ : sexp_option $tp$ >>
     | <:ctyp< $lid:nm$ : mutable $tp$ >>
     | <:ctyp< $lid:nm$ : $tp$ >> ->
-        let unrolled =
+      let unrolled =
           unroll_cnv_fp _loc <:expr< field_sexp >> (type_of_sexp _loc tp)
         in
         <:match_case<
@@ -933,7 +1009,9 @@ module Generate_of_sexp = struct
               | <:ctyp< sexp_option $_$ >>
               | <:ctyp< mutable sexp_option $_$ >>
               | <:ctyp< sexp_list $_$ >>
-              | <:ctyp< mutable sexp_list $_$ >> ->
+              | <:ctyp< mutable sexp_list $_$ >>
+              | <:ctyp< sexp_array $_$ >>
+              | <:ctyp< mutable sexp_array $_$ >> ->
                   bi_lst, <:patt< $lid:nm ^ "_value"$ >> :: good_patts
               | _ ->
                   has_nonopt_fields := true;
@@ -971,6 +1049,12 @@ module Generate_of_sexp = struct
                   match $lid:nm ^ "_value"$ with
                   [ None -> [] | Some v -> v ]
               >>
+          | <:ctyp< $lid:nm$ : sexp_array $_$ >> ->
+              <:rec_binding<
+                $lid:nm$ =
+                  match $lid:nm ^ "_value"$ with
+                  [ None -> [||] | Some v -> v ]
+              >>
           | <:ctyp< $lid:nm$ : $_$ >> ->
               <:rec_binding< $lid:nm$ = $lid:nm ^ "_value"$ >>
           | _ -> assert false  (* impossible *)
@@ -1101,7 +1185,9 @@ module Generate_of_sexp = struct
       is_alias_ref := true;
       type_of_sexp _loc tp
     in
-    let coll_args tp param = <:ctyp< $tp$ $param$ >> in
+    let coll_args tp param =
+      <:ctyp< $tp$ $Gen.drop_variance_annotations _loc param$ >>
+    in
     let full_type =
       List.fold_left ~f:coll_args ~init:<:ctyp< $lid:type_name$ >> tps
     in
