@@ -6,8 +6,8 @@
 
       Jane Street Holding, LLC
       Author: Markus Mottl
-      email: mmottl\@janestcapital.com
-      WWW: http://www.janestcapital.com/ocaml
+      email: mmottl\@janestreet.com
+      WWW: http://www.janestreet.com/ocaml
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -167,6 +167,8 @@ let output = output_mach
 
 (* Output of S-expressions to file *)
 
+(* CRv200912 sweeks for mmottl: Should this write to a temp and do an atomic move.
+ *)
 let save_of_output output_function file sexp =
   let oc = open_out file in
   try
